#include <stdio.h>

  /* Broj particija p(n) - Eulerova rekurzivna formula:
     p(n) = broj rastava zadanog prirodnog broja n
     u sumu nepadajucih (prirodnih) pribrojnika.
     Argument n smije biti cijeli broj.
     Dogovorno je p(0) = 1 i p(n) = 0, za n < 0.

     Nerekurzivna funkcija za p (po Eulerovoj formuli).
     Koristi globalno polje pbr za spremanje broja particija
     (kao zamjenu za rekurzivne pozive).

     Umjesto citanja broja n, ispisuje broj particija p(n),
     za sve brojeve n = 0, ..., 121.
     p(121) = najveci prikazivi broj particija u tipu int.
     Ispisuje i broj potrebnih vrijednosti iz polja pbr,
     samo za dani n i ukupno - za sve brojeve do tog n.
  */

    /* Globalno polje pbr, pbr[n] = broj particija od n. */

int pbr[122];

    /* Globalna varijabla zadnji_p sadrzi indeks zadnjeg izracunatog
       elementa u polju pbr. Inicijalizacija je zadnji_p = -1, sto
       znaci da niti jedan element u pbr jos nije postavljen. */

int zadnji_p = -1;


    /* Globalni brojac potrebnih vrijednosti iz globalnog polja pbr. */

unsigned int broj_potrebnih;


    /* Eulerova formula za broj particija p(n), za n >= 1:
         p(n) = suma_(m = 1)^(beskonacno)
                  (-1)^(m - 1) * ( p(n1_m) + p(n2_m) ),
       gdje su argumenti na desnoj strani (u sumi)
         n1_m = n - m * (3m - 1) / 2,
         n2_m = n - m * (3m + 1) / 2 = n1_m - m,
       uz p(0) = 1 i p(n) = 0, za n < 0.

       Implementacija: Uociti da suma ide najdalje do m = n.
       Nadalje, za argumente vrijedi n1_m > n2_m = n1_m - m.
       U uzlaznoj petlji po m, prvo racunamo veci argument n1_m.
       Kad dobijemo da je n1_m < 0, prekidamo petlju po m.

       Umjesto gornje izravne formule za n1_m, koristimo vezu
       izmedju dva susjedna clana (po m):
         n1_m = n1_(m - 1) - 3*m + 2, za m >= 1,
       uz pocetak n1_0 = n. To malo ubrzava racunanje. */


    /* Nerekurzivna funkcija za broj particija p(n).
       Koristi globalno polje pbr za spremanje broja particija
       (tako da se izbjegnu nepotrebni rekurzivni pozivi).

       Napomena: Funkcija OCEKUJE da je zadnji_p = n - 1, tj. da
       su VEC izracunati svi brojevi particija p(k), za k < n.
       Funkcija racuna p(n), sprema ga u pbr i povecava zadnji_p.
       Drugim rijecima, ocekuje se pozivanje u petlji, pocev od 0,
       do nekog n, s korakom 1. */

int p(int n)
{
    int broj = 0, m, n1_m, n2_m, pribr;

        /* Provjera argumenata i korektnosti poziva.
           Ocekujemo da je n >= 0 i zadnji_p = n - 1. */
    if (n < 0 || zadnji_p != n - 1) return 0;

        /* Prvi poziv za n = 0. */
    if (n == 0)
        broj = 1;

    else {    /* n > 0. */

            /* Pocetni veci argument n1_0 = n, za m = 0. */
        n1_m = n;

            /* Suma iz Eulerove formule - uzlazna petlja po m.
               Vrijednosti funkcije p za argumente n1_m, n2_m
               (oba su < n) uzimamo iz polja pbr (vec su izracunate).
               Pazimo da su argumenti >= 0 (nema nepotrebnih poziva). */

        for (m = 1; m <= n; ++m) {

                /* Prvo racunamo veci argument n1_m iz prethodnog. */
            n1_m = n1_m - 3*m + 2;

                /* Kad dobijemo n1_m < 0, prekidamo petlju po m. */
            if (n1_m < 0) break;

                /* Procitaj p(n1_m) iz polja pbr. */
            pribr = pbr[n1_m];
                /* Povecaj globalni brojac potrebnih vrijenosti iz pbr. */
            ++broj_potrebnih;

                /* Manji argument n2_m iz n1_m. */
            n2_m = n1_m - m;

                /* Provjera manjeg argumenta, n2_m >= 0. */
            if (n2_m >= 0) {
                    /* Procitaj p(n2_m) iz polja pbr. */
                pribr += pbr[n2_m];
                    /* Povecaj globalni brojac potrebnih vrijenosti iz pbr. */
                ++broj_potrebnih;
            }

                /* Provjera predznaka (parnosti od m) - ovisno o tome,
                   broju treba dodati ili oduzeti pribrojnik pribr. */
            if (m % 2 == 1)    /* (m - 1) je paran. */
                broj += pribr;
            else
                broj -= pribr;
        }
    }

        /* Postavi broj = p(n) na mjesto pbr[n], povecaj zadnji_p. */
    pbr[n] = broj;
    zadnji_p = n;

    return broj;
}

    /* Glavni program za tablicu broja particija. */

int main(void)
{
    int n, broj;
    unsigned int uk_potrebno = 0u;

        /* Zaglavlje tablice. */
    printf("\n");
    printf("     Broj particija      Potrebno    Ukupno\n");
    printf(" --------------------------------------------\n");

    for (n = 0; n <= 121; ++n ) {

            /* Inicijalizacija globalnog brojaca potrebnih
               vrijednosti iz polja pbr (samo za ovaj n). */
        broj_potrebnih = 0u;

            /* Poziv funkcije p. */
        broj = p(n);
        uk_potrebno += broj_potrebnih;

            /* Redak tablice. */
        printf("   p(%3d) = %10d   %8u  %8u\n",
               n, broj, broj_potrebnih, uk_potrebno);
    }

    return 0;
}
